@@ -1,9 +1,15 @@
 import datetime
-import json
 import os
 import signal
 import sys
 import warnings
+from openai.validators import (
+    write_out_file,
+    apply_necessary_remediation,
+    apply_optional_remediation,
+    read_any_format,
+    get_validators,
+)
 
 import openai
 
@@ -138,8 +144,14 @@ class Completion:
         if args.n is not None and args.n > 1 and args.stream:
             raise ValueError("Can't stream completions with n>1 with the current CLI")
 
+        if args.engine and args.model:
+            warnings.warn(
+                "In most cases, you should not be specifying both engine and model."
+            )
+
         resp = openai.Completion.create(
             engine=args.engine,
+            model=args.model,
             n=args.n,
             max_tokens=args.max_tokens,
             logprobs=args.logprobs,
@@ -215,12 +227,48 @@ class FineTune:
         print(resp)
 
     @classmethod
-    def _get_or_upload(cls, file):
+    def _get_or_upload(cls, file, check_if_file_exists=True):
         try:
             openai.File.retrieve(file)
         except openai.error.InvalidRequestError as e:
             if e.http_status == 404 and os.path.isfile(file):
-                resp = openai.File.create(file=open(file), purpose="fine-tune")
+                matching_files = openai.File.find_matching_files(
+                    file=open(file), purpose="fine-tune"
+                )
+                if len(matching_files) > 0 and check_if_file_exists:
+                    file_ids = [f["id"] for f in matching_files]
+                    sys.stdout.write(
+                        "Found potentially duplicated files with name '{name}', purpose 'fine-tune' and size {size} bytes\n".format(
+                            name=matching_files[0]["filename"],
+                            size=matching_files[0]["bytes"],
+                        )
+                    )
+                    sys.stdout.write("\n".join(file_ids))
+                    while True:
+                        sys.stdout.write(
+                            "\nEnter file ID to reuse an already uploaded file, or an empty string to upload this file anyway: "
+                        )
+                        inp = sys.stdin.readline().strip()
+                        if inp in file_ids:
+                            sys.stdout.write(
+                                "Using your file {file}: {id}\n".format(
+                                    file=file, id=inp
+                                )
+                            )
+                            return inp
+                        elif inp == "":
+                            break
+                        else:
+                            sys.stdout.write(
+                                "File id '{id}' is not among the IDs of the potentially duplicated files\n".format(
+                                    id=inp
+                                )
+                            )
+
+                resp = openai.File.create(
+                    file=open(file),
+                    purpose="fine-tune",
+                )
                 sys.stdout.write(
                     "Uploaded file from {file}: {id}\n".format(file=file, id=resp["id"])
                 )
@@ -230,64 +278,79 @@ class FineTune:
     @classmethod
     def create(cls, args):
         create_args = {
-            "training_file": cls._get_or_upload(args.training_file),
+            "training_file": cls._get_or_upload(
+                args.training_file, args.check_if_files_exist
+            ),
         }
         if args.validation_file:
-            create_args["validation_file"] = cls._get_or_upload(args.validation_file)
-        if args.model:
-            create_args["model"] = args.model
-        if args.hparams:
-            try:
-                hparams = json.loads(args.hparams)
-            except json.decoder.JSONDecodeError:
-                sys.stderr.write(
-                    "--hparams must be JSON decodable and match the hyperparameter arguments of the API"
-                )
-                sys.exit(1)
-            create_args.update(hparams)
+            create_args["validation_file"] = cls._get_or_upload(
+                args.validation_file, args.check_if_files_exist
+            )
+
+        for hparam in (
+            "model",
+            "n_epochs",
+            "batch_size",
+            "learning_rate_multiplier",
+            "prompt_loss_weight",
+            "use_packing",
+            "compute_classification_metrics",
+            "classification_n_classes",
+            "classification_positive_class",
+            "classification_betas",
+        ):
+            attr = getattr(args, hparam)
+            if attr is not None:
+                create_args[hparam] = attr
 
         resp = openai.FineTune.create(**create_args)
 
-        if args.no_wait:
+        if args.no_follow:
             print(resp)
             return
 
         sys.stdout.write(
-            "Created job: {job_id}\n"
-            "Streaming events until the job is complete...\n\n"
-            "(Ctrl-C will interrupt the stream, but not cancel the job)\n".format(
+            "Created fine-tune: {job_id}\n"
+            "Streaming events until fine-tuning is complete...\n\n"
+            "(Ctrl-C will interrupt the stream, but not cancel the fine-tune)\n".format(
                 job_id=resp["id"]
             )
         )
         cls._stream_events(resp["id"])
 
-        resp = openai.FineTune.retrieve(id=resp["id"])
-        status = resp["status"]
-        sys.stdout.write("\nJob complete! Status: {status}".format(status=status))
-        if status == "succeeded":
-            sys.stdout.write(" 🎉")
-            sys.stdout.write(
-                "\nTry out your fine-tuned model: {model}\n"
-                "(Pass this as the model parameter to a completion request)".format(
-                    model=resp["fine_tuned_model"]
-                )
-            )
-            # TODO(rachel): Print instructions on how to use the model here.
-        elif status == "failed":
-            sys.stdout.write("\nPlease contact support@openai.com for assistance.")
-        sys.stdout.write("\n")
-
     @classmethod
     def get(cls, args):
         resp = openai.FineTune.retrieve(id=args.id)
         print(resp)
+        print(resp["result_files"][0])
+
+    @classmethod
+    def results(cls, args):
+        fine_tune = openai.FineTune.retrieve(id=args.id)
+        if "result_files" not in fine_tune or len(fine_tune["result_files"]) == 0:
+            raise openai.error.InvalidRequestError(
+                f"No results file available for fine-tune {args.id}", "id"
+            )
+        result_file = openai.FineTune.retrieve(id=args.id)["result_files"][0]
+        resp = openai.File.download(id=result_file["id"])
+        print(resp.decode("utf-8"))
 
     @classmethod
     def events(cls, args):
-        if not args.stream:
-            resp = openai.FineTune.list_events(id=args.id)  # type: ignore
-            print(resp)
-            return
+        if args.stream:
+            raise openai.error.OpenAIError(
+                message=(
+                    "The --stream parameter is deprecated, use fine_tunes.follow "
+                    "instead:\n\n"
+                    "  openai api fine_tunes.follow -i {id}\n".format(id=args.id)
+                ),
+            )
+
+        resp = openai.FineTune.list_events(id=args.id)  # type: ignore
+        print(resp)
+
+    @classmethod
+    def follow(cls, args):
         cls._stream_events(args.id)
 
     @classmethod
@@ -295,9 +358,11 @@ class FineTune:
         def signal_handler(sig, frame):
             status = openai.FineTune.retrieve(job_id).status
             sys.stdout.write(
-                "\nStream interrupted. Job is still {status}. "
-                "To cancel your job, run:\n"
-                "`openai api fine_tunes.cancel -i {job_id}`\n".format(
+                "\nStream interrupted. Job is still {status}.\n"
+                "To resume the stream, run:\n\n"
+                "  openai api fine_tunes.follow -i {job_id}\n\n"
+                "To cancel your job, run:\n\n"
+                "  openai api fine_tunes.cancel -i {job_id}\n\n".format(
                     status=status, job_id=job_id
                 )
             )
@@ -307,24 +372,109 @@ class FineTune:
 
         events = openai.FineTune.stream_events(job_id)
         # TODO(rachel): Add a nifty spinner here.
-        for event in events:
+        try:
+            for event in events:
+                sys.stdout.write(
+                    "[%s] %s"
+                    % (
+                        datetime.datetime.fromtimestamp(event["created_at"]),
+                        event["message"],
+                    )
+                )
+                sys.stdout.write("\n")
+                sys.stdout.flush()
+        except Exception:
             sys.stdout.write(
-                "[%s] %s"
-                % (
-                    datetime.datetime.fromtimestamp(event["created_at"]),
-                    event["message"],
+                "\nStream interrupted (client disconnected).\n"
+                "To resume the stream, run:\n\n"
+                "  openai api fine_tunes.follow -i {job_id}\n\n".format(job_id=job_id)
+            )
+            return
+
+        resp = openai.FineTune.retrieve(id=job_id)
+        status = resp["status"]
+        if status == "succeeded":
+            sys.stdout.write("\nJob complete! Status: succeeded 🎉")
+            sys.stdout.write(
+                "\nTry out your fine-tuned model:\n\n"
+                "openai api completions.create -m {model} -p <YOUR_PROMPT>".format(
+                    model=resp["fine_tuned_model"]
                 )
             )
-            sys.stdout.write("\n")
-            sys.stdout.flush()
+        elif status == "failed":
+            sys.stdout.write(
+                "\nJob failed. Please contact support@openai.com if you need assistance."
+            )
+        sys.stdout.write("\n")
 
     @classmethod
     def cancel(cls, args):
         resp = openai.FineTune.cancel(id=args.id)
         print(resp)
 
+    @classmethod
+    def prepare_data(cls, args):
 
-def register(parser):
+        sys.stdout.write("Analyzing...\n")
+        fname = args.file
+        df, remediation = read_any_format(fname)
+        apply_necessary_remediation(None, remediation)
+
+        validators = get_validators()
+
+        optional_remediations = []
+        if remediation is not None:
+            optional_remediations.append(remediation)
+        for validator in validators:
+            remediation = validator(df)
+            if remediation is not None:
+                optional_remediations.append(remediation)
+                df = apply_necessary_remediation(df, remediation)
+
+        any_optional_or_necessary_remediations = any(
+            [
+                remediation
+                for remediation in optional_remediations
+                if remediation.optional_msg is not None
+                or remediation.necessary_msg is not None
+            ]
+        )
+
+        if any_optional_or_necessary_remediations:
+            sys.stdout.write(
+                "\n\nBased on the analysis we will perform the following actions:\n"
+            )
+
+            for remediation in optional_remediations:
+                df = apply_optional_remediation(df, remediation)
+        else:
+            sys.stdout.write("\n\nNo remediations found.\n")
+
+        write_out_file(df, fname, any_optional_or_necessary_remediations)
+
+
+def tools_register(parser):
+    subparsers = parser.add_subparsers(
+        title="Tools", help="Convenience client side tools"
+    )
+
+    def help(args):
+        parser.print_help()
+
+    parser.set_defaults(func=help)
+
+    sub = subparsers.add_parser("fine_tunes.prepare_data")
+    sub.add_argument(
+        "-f",
+        "--file",
+        required=True,
+        help="JSONL, JSON, CSV, TSV, TXT or XLSX file containing prompt-completion examples to be analyzed."
+        "This should be the local file path.",
+    )
+    sub.set_defaults(func=FineTune.prepare_data)
+
+
+def api_register(parser):
     # Engine management
     subparsers = parser.add_subparsers(help="All API subcommands")
 
@@ -422,7 +572,16 @@ Mutually exclusive with `top_p`.""",
 
     # Completions
     sub = subparsers.add_parser("completions.create")
-    sub.add_argument("-e", "--engine", required=True, help="The engine to use")
+    sub.add_argument(
+        "-e",
+        "--engine",
+        help="The engine to use. See https://beta.openai.com/docs/engines for more about what engines are available.",
+    )
+    sub.add_argument(
+        "-m",
+        "--model",
+        help="The model to use. At most one of `engine` or `model` should be specified.",
+    )
     sub.add_argument(
         "--stream", help="Stream tokens as they're ready.", action="store_true"
     )
@@ -529,34 +688,119 @@ Mutually exclusive with `top_p`.""",
         "be the ID of a file uploaded through the OpenAI API (e.g. file-abcde12345) "
         "or a local file path.",
     )
+    sub.add_argument(
+        "--no_check_if_files_exist",
+        dest="check_if_files_exist",
+        action="store_false",
+        help="If this argument is set and training_file or validation_file are file paths, immediately upload them. If this argument is not set, check if they may be duplicates of already uploaded files before uploading, based on file name and file size.",
+    )
     sub.add_argument(
         "-m",
         "--model",
         help="The model to start fine-tuning from",
     )
     sub.add_argument(
-        "--no_wait",
+        "--no_follow",
         action="store_true",
-        help="If set, returns immediately after creating the job. Otherwise, waits for the job to complete.",
+        help="If set, returns immediately after creating the job. Otherwise, streams events and waits for the job to complete.",
+    )
+    sub.add_argument(
+        "--n_epochs",
+        type=int,
+        help="The number of epochs to train the model for. An epoch refers to one "
+        "full cycle through the training dataset.",
+    )
+    sub.add_argument(
+        "--batch_size",
+        type=int,
+        help="The batch size to use for training. The batch size is the number of "
+        "training examples used to train a single forward and backward pass.",
+    )
+    sub.add_argument(
+        "--learning_rate_multiplier",
+        type=float,
+        help="The learning rate multiplier to use for training. The fine-tuning "
+        "learning rate is determined by the original learning rate used for "
+        "pretraining multiplied by this value",
+    )
+    sub.add_argument(
+        "--use_packing",
+        action="store_true",
+        dest="use_packing",
+        help="On classification tasks, we recommend not setting this flag. "
+        "On all other tasks, we recommend setting it. "
+        "When set, we pack as many prompt-completion pairs as possible into each "
+        "training example. This greatly increases the speed of a fine-tuning job, "
+        "often without negatively affecting model performance.",
+    )
+    sub.add_argument(
+        "--no_packing",
+        action="store_false",
+        dest="use_packing",
+        help="Disables the packing flag (see --use_packing for description)",
+    )
+    sub.set_defaults(use_packing=None)
+    sub.add_argument(
+        "--prompt_loss_weight",
+        type=float,
+        help="The weight to use for the prompt loss. The optimum value here depends "
+        "depends on your use case. This determines how much the model prioritizes "
+        "learning from prompt tokens vs learning from completion tokens",
+    )
+    sub.add_argument(
+        "--compute_classification_metrics",
+        action="store_true",
+        help="If set, we calculate classification-specific metrics such as accuracy "
+        "and F-1 score using the validation set at the end of every epoch.",
+    )
+    sub.set_defaults(compute_classification_metrics=None)
+    sub.add_argument(
+        "--classification_n_classes",
+        type=int,
+        help="The number of classes in a classification task. This parameter is "
+        "required for multiclass classification",
+    )
+    sub.add_argument(
+        "--classification_positive_class",
+        help="The positive class in binary classification. This parameter is needed "
+        "to generate precision, recall and F-1 metrics when doing binary "
+        "classification",
+    )
+    sub.add_argument(
+        "--classification_betas",
+        type=float,
+        nargs="+",
+        help="If this is provided, we calculate F-beta scores at the specified beta "
+        "values. The F-beta score is a generalization of F-1 score. This is only "
+        "used for binary classification.",
     )
-    sub.add_argument("-p", "--hparams", help="Hyperparameter JSON")
     sub.set_defaults(func=FineTune.create)
 
     sub = subparsers.add_parser("fine_tunes.get")
     sub.add_argument("-i", "--id", required=True, help="The id of the fine-tune job")
     sub.set_defaults(func=FineTune.get)
 
+    sub = subparsers.add_parser("fine_tunes.results")
+    sub.add_argument("-i", "--id", required=True, help="The id of the fine-tune job")
+    sub.set_defaults(func=FineTune.results)
+
     sub = subparsers.add_parser("fine_tunes.events")
     sub.add_argument("-i", "--id", required=True, help="The id of the fine-tune job")
+
+    # TODO(rachel): Remove this in 1.0
     sub.add_argument(
         "-s",
         "--stream",
         action="store_true",
-        help="If set, events will be streamed until the job is done. Otherwise, "
+        help="[DEPRECATED] If set, events will be streamed until the job is done. Otherwise, "
         "displays the event history to date.",
     )
     sub.set_defaults(func=FineTune.events)
 
+    sub = subparsers.add_parser("fine_tunes.follow")
+    sub.add_argument("-i", "--id", required=True, help="The id of the fine-tune job")
+    sub.set_defaults(func=FineTune.follow)
+
     sub = subparsers.add_parser("fine_tunes.cancel")
     sub.add_argument("-i", "--id", required=True, help="The id of the fine-tune job")
     sub.set_defaults(func=FineTune.cancel)
